/* src/fichiers.c - Lecture/criture des donnes
 * Copyright (C) 2002-2004 Inter System
 *
 * contact: Progs@Inter-System.Net
 *          Cesar@Inter-System.Net
 *          kouak@kouak.org
 * site web: http://coderz.inter-system.net
 *
 * Services pour serveur IRC. Support sur IrcProgs et IrCoderZ
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * $Id: fichiers.c,v 1.100 2004/03/31 11:38:51 romexzf Exp $
 */

#include "main.h"
#include <errno.h>
#include "fichiers.h"
#include "outils.h"
#include "config.h"
#include "hash.h"
#include "debug.h"
#include "cs_cmds.h"
#include "add_info.h"
#include "admin_user.h"
#include "admin_chan.h"
#ifdef HAVE_VOTE
#include "vote.h"
#endif
#ifdef USE_WELCOMESERV
#include "welcome.h"
#endif

static size_t ulen = 0;
static size_t clen = 0;

int load_users(void)
{
	FILE *fp, *fp1;
	char buff[100];
	int adm = 0, fileversion = 1, stats = 0;
	anUser *u = NULL;
	anAccess *a;

	if(!ulen) ulen = strlen(DBDIR "/usr/");

	if(!(fp = fopen(USERS_FILE, "r")))
	{
		ConfFlag |= CF_PREMIERE;
		return 0;
	}
	while(fgets(buff, sizeof buff, fp))
	{
		char buff1[500] = DBDIR "/usr/";
		stats++;
		strip_newline(buff);
		conv_char(buff, buff1 + ulen);
		if(!(fp1 = fopen(buff1, "r")))
		{
			Debug(W_TTY, "DB-load: Impossible d'ouvrir le fichier '%s' pour '%s': %s", buff1, buff, strerror(errno));
			continue;
		}

		while(fgets(buff1, sizeof buff1, fp1))
		{
			char *field = NULL;
			strip_newline(buff1);
			if((field = strchr(buff1, ' '))) *field++ = 0;
			else continue;

			if(!strcmp(buff1, "VERSION")) fileversion = atoi(field);
			else if(!strcmp(buff1, "NICK"))
			{ /* NICK <user> <pass> <lvl> <TS> <flag> <vote> <mail> <lastlogin> <cantregchan> <lang> */
				char *nick, *pass, *level, *suspendu = NULL, *mail, *lastseen = NULL, *voted = NULL, *last = NULL;
				char *cantregchan = NULL, *lang = NULL;
				int flag = 0;

				nick = strtok(field, " ");
				pass = strtok(NULL, " ");
				level = strtok(NULL, " ");

				suspendu = strtok(NULL, " ");
				lastseen = strtok(NULL, " ");
				flag = atoi(strtok(NULL, " "));
				voted = strtok(NULL, " ");
				mail = strtok(NULL, " ");
				last = strtok(NULL, " ");
				cantregchan = strtok(NULL, " ");
				if(!cantregchan)
				{
					Debug(W_TTY, "Erreur lors de la lecture de la DB pour %s, item incomplet", nick);
					continue;
				}
				if(fileversion >= 20) lang = strtok(NULL, " ");

				u = add_regnick(nick, pass, lastseen ? atol(lastseen) : 0,
					atoi(level), flag, mail, (voted && is_num(voted)) ? atoi(voted) : 0);

				if(*last != '0') str_dup(&u->lastlogin, last);
				if((u->cantregchan = strtol(cantregchan, NULL, 10)) > 0 && u->cantregchan < CurrentTS) u->cantregchan = -1;
				if(u->level >= ADMINLEVEL) adm = 1; /* at least one admin in DB ? */
				if(lang) u->lang = lang_isloaded(lang); /* try to load language 'lang' */
				if(!u->lang) u->lang = DefaultLang; /* if not more loaded or old DB, then Default */
			}

			else if(!strcmp(buff1, "SUSPEND") && *field && fileversion >= 16)
			{ /* UserName TimeStamp :raison */
				char *from, *expire, *raison, *debut;
				raison = strchr(field, ':');
				from = strtok(field, " ");
				expire = strtok(NULL, " ");
				debut = strtok(NULL, " ");
				do_user_suspend(u, ++raison, from, strtol(expire, NULL, 10), strtol(debut, NULL, 10));
			}

#ifdef USE_MEMOSERV
			else if(!strcmp(buff1, "MEMO"))
			{ /* MEMO <from> <DateTS> <flag> :message */
				char *de, *date, *read = NULL, *tmp = strchr(field, ':');

				de = strtok(field, " ");
				date = strtok(NULL, " ");
				read = strtok(NULL, " ");
				add_memo(u, de, atoi(date), ++tmp, read ? atol(read) : 0);
			}
#endif
			else if(!strcmp(buff1, "ACCESS"))
			{ /* ACCESS <#> <level> <flag> <lastseenTS> :info */
				int flaga = 0;
				char *tmp = strchr(field, ':'), *chan, *level, *lastseen = NULL;

				chan = strtok(field, " ");
				level = strtok(NULL, " ");
				flaga = atoi(strtok(NULL, " "));
				lastseen = strtok(NULL, " ");
				a = add_access(u, chan, atoi(level), flaga, lastseen ? atol(lastseen) : 0);
				if(a && tmp && *++tmp) str_dup(&a->info, tmp);
			}
		}
		fclose(fp1);
	}
	fclose(fp);
	if(!adm) Debug(W_TTY, "Aucun Admin localis dans la Base de Donnes!!");
	return stats;
}

int write_userlist(void)
{
	FILE *fp = fopen(USERS_FILE, "w");
	anUser *user;
	int i = 0;

	if(!fp)	return Debug(W_MAX|W_WARN, "DB-W: Impossible d'crire le fichier "USERS_FILE": %s", strerror(errno));

	for(;i < USERHASHSIZE;i++) for(user = user_tab[i];user;user = user->next)
		fputs(user->nick, fp), fputc('\n', fp);

	fclose(fp);
	return 0;
}

int write_anuser(anUser *u)
{
#ifdef USE_MEMOSERV
	aMemo *memo = u->memohead;
#endif
	anAccess *a = u->accesshead;
 	FILE *fp;
 	char buff[140] = DBDIR "/usr/";

 	conv_char(u->nick, buff + ulen);
	if(!(fp = fopen(buff, "w")))
		return Debug(W_MAX|W_WARN, "DB-W: Impossible d'crire le fichier '%s' pour %s: %s", buff, u->nick, strerror(errno));

	fprintf(fp, "VERSION %d\n", USRS_VERSION);
#ifdef HAVE_VOTE
	fprintf(fp, "NICK %s %s %d 0 %lu %d %d %s %s %ld %s\n", u->nick, u->passwd, u->level,
		u->lastseen, u->flag, u->vote, u->mail, u->lastlogin ? u->lastlogin : "0", u->cantregchan, u->lang->langue);
#else
	fprintf(fp, "NICK %s %s %d 0 %lu %d 0 %s %s %ld %s\n", u->nick, u->passwd,
		u->level, u->lastseen, u->flag, u->mail, u->lastlogin ? u->lastlogin : "0", u->cantregchan, u->lang->langue);
#endif
	if(u->suspend) fprintf(fp, "SUSPEND %s %ld %ld :%s\n", u->suspend->from,
			u->suspend->expire, u->suspend->debut, u->suspend->raison);

#ifdef USE_MEMOSERV
	for(;memo;memo = memo->next) fprintf(fp, "MEMO %s %d %d :%s\n", memo->de, memo->date, memo->read, memo->message);
#endif

    for(;a;a = a->next)
    	fprintf(fp, "ACCESS %s %d %d %lu :%s\n", a->c->chan, a->level,
    		a->flag, a->lastseen == 1 ? CurrentTS : a->lastseen, NONE(a->info));

    fclose(fp);
    return 0;
}

int write_chanlist(void)
{
	FILE *fp = fopen(CHANS_FILE, "w");
	int i = 0;
	aChan *chan;

	if(!fp)	return Debug(W_MAX|W_WARN, "DB-W: Impossible d'crire le fichier "CHANS_FILE": %s", strerror(errno));
	for(;i < CHANHASHSIZE;i++) for(chan = chan_tab[i];chan; chan = chan->next)
		fputs(chan->chan, fp), fputc('\n', fp);

	fclose(fp);
	return 0;
}

int write_achan(aChan *chan)
{
	FILE *fchan;
	char buff[230] = DBDIR "/chans/";

	conv_char(chan->chan, buff + clen);
	if(!(fchan = fopen(buff, "w")))
		return Debug(W_MAX|W_WARN, "DB-W: Impossible d'crire le fichier '%s' pour %s: %s", buff, chan->chan, strerror(errno));

	if(chan->defmodes.modes) fprintf(fchan, "DEFMODES %s\n", GetCModes(chan->defmodes));
	if(*chan->deftopic) fprintf(fchan, "DEFTOPIC %s\n", chan->deftopic);
	fprintf(fchan, "DESCRIPTION %s\n", chan->description);
	if(*chan->welcome) fprintf(fchan, "WELCOME %s\n", chan->welcome);
	fprintf(fchan, "OPTIONS %d %d %d %d %lu %d %d\n", (chan->flag & ~C_JOINED), chan->banlevel,
		chan->bantype, chan->cml, chan->bantime, chan->limit_min, chan->limit_inc);
	if(chan->motd) fprintf(fchan, "MOTD %s\n", chan->motd);
	fprintf(fchan, "URL %s %lu\n", chan->url, chan->creation_time);
	if(chan->suspend)
		fprintf(fchan, "SUSPEND %s %ld %ld :%s\n", chan->suspend->from, chan->suspend->expire, chan->suspend->debut, chan->suspend->raison);

	fclose(fchan);
	return 0;
}

int write_bans(aChan *chan)
{
	FILE *fchan;
	aBan *ban = chan->banhead;
	char buff[230] = DBDIR "/bans/";

	conv_char(chan->chan, buff + clen - 1);
	if(!ban) /* au lieu d'crire un fichier pour keud, on peut kiter direct, et supprimer le fichier*/
	{		      /* si il existe pour faire le mnage :p --Progs*/
		remove(buff);
		return 0;
	}

	if(!(fchan = fopen(buff, "w")))
		return Debug(W_MAX, "DB-W: Impossible d'crire le fichier '%s' pour %s: %s", buff, chan->chan, strerror(errno));
	for(;ban;ban = ban->next)
		fprintf(fchan, "%s %s %lu %lu %d :%s\n", ban->mask, ban->de, ban->debut, ban->fin, ban->level, ban->raison);

	fclose(fchan);
	return 0;
}

int count_char(const char *buf, int c)
{
	int count = 0;
	while(*buf) if(*buf++ == (char) c) count++;
	return count;
}

int load_chans(void)
{
	FILE *fp, *fp1;
	char buff[250];
	int stats = 0;

	if(!clen) clen = strlen(DBDIR "/chans/");

	if(!(fp = fopen(CHANS_FILE, "r")))
	{
		ConfFlag |= CF_PREMIERE;
		return 0;
	}
	while(fgets(buff, sizeof buff, fp))
	{
		char *chan = buff, nchan[2 * REGCHANLEN] = {0}, buff1[500] = DBDIR "/chans/";
		aChan *cc;

		stats++;
		strip_newline(buff);
		conv_char(chan, nchan);
		strcpy(buff1 + clen, nchan);
		if(!(fp1 = fopen(buff1, "r")))
		{
			Debug(W_TTY, "DB-load: Impossible d'ouvrir le fichier '%s' pour '%s': %s", buff1, buff, strerror(errno));
			continue;
		}

		if(!(cc = add_chan(chan, "")))
		{
			Debug(W_TTY, "Erreur lors du chargement de %s", chan);
			break;
		}

		while(fgets(buff1, sizeof buff1, fp1))
		{
			char *field = strchr(buff1, ' ');
			strip_newline(buff1);
			if(field) *field++ = 0;
			if(!field || !*field) continue;

			if(!strcmp(buff1, "DEFMODES"))
			{
				char *lim = strchr(field, ' '), *key = NULL;
				if(lim)
				{
					*lim++ = 0;
					if((key = strchr(lim, ' '))) *key++ = 0;
				}
				string2scmode(&cc->defmodes, field, !key && strchr(field, 'k') ? lim : key, lim);
			}
			else if(!strcmp(buff1, "DEFTOPIC")) Strncpy(cc->deftopic, field, TOPICLEN);
			else if(!strcmp(buff1, "DESCRIPTION")) Strncpy(cc->description, field, DESCRIPTIONLEN);
			else if(!strcmp(buff1, "WELCOME")) Strncpy(cc->welcome, field, TOPICLEN);
			else if(!strcmp(buff1, "OPTIONS"))
			{
				int flag = 0, bantype = 0;
				unsigned long bandefault = 0;
				if(count_char(field, ' ') > 6)/* vieux format*/
				{
					int setwelcome = 0, strictop = 0, locktopic = 0, nobans = 0, noops = 0, autoinvite = 0;
					sscanf(field, "%d %d %d %d %d %d %d %d %d %lu", &setwelcome, &strictop,	&locktopic,
						&nobans, &noops, &cc->banlevel, &bantype, &autoinvite, &cc->cml, &bandefault);
					if(setwelcome) flag |= C_SETWELCOME;
					if(nobans) flag |= C_NOBANS;
					if(noops) flag |= C_NOOPS;
					if(locktopic) flag |= C_LOCKTOPIC;
					if(strictop) flag |= C_STRICTOP;
					if(autoinvite) flag |= C_AUTOINVITE;
				}
				else sscanf(field, "%d %d %d %d %lu %d %d", &flag, &cc->banlevel, &bantype, &cc->cml,
					&bandefault, &cc->limit_min, &cc->limit_inc);

				cc->flag = (flag & ~C_JOINED);
				cc->bantype = (bantype > 4 || bantype < 1) ? 1 : bantype;
				cc->bantime = (time_t) bandefault;
			}
			else if(!strcmp(buff1, "MOTD")) str_dup(&cc->motd, field);
			else if(!strcmp(buff1, "URL")) /* on peut ainsi supprimer une ligne des DBs chans*/
			{
				char *creation_time = strchr(field, ' ');
				if(creation_time) *creation_time++ = 0;
				Strncpy(cc->url, field, 80);
				cc->creation_time = (creation_time && *creation_time) ? strtoul(creation_time, NULL, 10) : 0;
			}
			else if(!strcmp(buff1, "SUSPEND"))
			{ /* UserName TimeStamp :raison */
				char *from, *expire, *raison, *debut;
				raison = strchr(field, ':');
				from = strtok(field, " ");
				expire = strtok(NULL, " ");
				debut = strtok(NULL, " ");
				do_chan_suspend(cc, ++raison, from, strtol(expire, NULL, 10), strtol(debut, NULL, 10));
			}
		}
		fclose(fp1);
		/* load bans */
		snprintf(buff1, sizeof buff1, DBDIR "/bans/%s", nchan);
		if(!(fp1 = fopen(buff1, "r"))) continue;
		while(fgets(buff1, sizeof buff1, fp1))
		{
			char *mask = NULL, *de = NULL, *level = NULL, *tmp;
			time_t debut, fin;
			strip_newline(buff1);
			tmp = strchr(buff1, ':');
			mask = strtok(buff1, " ");
			de = strtok(NULL, " ");
			debut = strtol(strtok(NULL, " "), NULL, 10);
			fin = strtol(strtok(NULL, " "), NULL, 10);
			if(fin && CurrentTS >= fin) continue;
			level = strtok(NULL, " ");
			add_ban(cc, mask, ++tmp, de, debut, fin ? fin - debut : 0, atoi(level));
		}
		fclose(fp1);
	}
	fclose(fp);
	return stats;
}

void write_badchans(void)
{
	FILE *fp;
	struct badchans *badchan = badchead;

	if(!(fp = fopen(BADCHANS_FILE, "w"))) return;

	for(; badchan; badchan = badchan->next)
		fprintf(fp, "%s :%s\n", badchan->chaine, badchan->raison);

	fclose(fp);
	return;
}

int load_badchans(void)
{
   char buff[300], *chaine, *raison = NULL;
   int stats = 0;
   FILE *fp = fopen(BADCHANS_FILE, "r");

   if(!fp) return Debug(W_TTY, "DB-load: Erreur lors de la lecture des badchans: %s", strerror(errno));
   while(fgets(buff, sizeof buff, fp))
   {
	   strip_newline(buff);
	   if((raison = strchr(buff, ':'))) *raison++ = '\0';
	   chaine = strtok(buff, " ");
	   add_abadchan(chaine, raison);
   }
   fclose(fp);
   return stats;
}

void write_badnicks(void)
{
	FILE *fp = fopen(BADNICKS_FILE, "w");
	struct badnicks *badnick = badnhead;

	if(!fp) return;

	for(;badnick;badnick = badnick->next)
		fprintf(fp, "%s %d\n", badnick->chaine, badnick->flag);

	fclose(fp);
	return;
}

int load_badnicks(void)
{
	char buff[100], *nick, *flag = NULL;
	int stats = 0;
	FILE *fp = fopen(BADNICKS_FILE, "r");

	if(!fp) return Debug(W_TTY, "DB-load: Erreur lors de la lecture des badnicks: %s", strerror(errno));

	while(fgets(buff, sizeof buff, fp))
	{
		int f = 0;
		strip_newline(buff);
		nick = strtok(buff, " ");
		flag = strtok(NULL, " ");
		if(flag && is_num(flag) && atoi(flag)) f |= CF_NOREG;
		add_abadnick(buff, f);
		stats++;
	}
	fclose(fp);
	return stats;
}

void write_cmds(void)
{
	FILE *fp = fopen(CMDS_FILE, "w");
	int i = 0;
	aHashCmd *cmd;
	if(!fp) return;

	for(;i < CMDHASHSIZE;++i) for(cmd = cmd_hash[i];cmd;cmd = cmd->next)
	{
		if(*cmd->name != '\001') fprintf(fp, "%s %s %d %d %d\n",
		  *cmd->name ? cmd->name : cmd->corename, cmd->corename, cmd->level, cmd->flag, cmd->used);
	}
	fclose(fp);
}

int load_cmds(void)
{
	char buff[80], name[CMDLEN + 1], corename[CMDLEN + 1], dir[50] = "syntax/";
	int i = 0;
	aHashCmd *cmd;
	FILE *f, *fp = fopen(CMDS_FILE, "r");

	if(!fp) return Debug(W_TTY, "DB-load: Erreur lors de la lecture des commandes: %s", strerror(errno));
	while(fgets(buff, sizeof buff, fp))
	{
		int level = 0, flag = 0, used = 0;
		sscanf(buff, "%14s %14s %d %d %d", name, corename, &level, &flag, &used);

		if(!(cmd = FindCommand(corename)))/* recherche sur corename, mais c'est au load*/
		{							/* donc corename == name*/
			Debug(W_TTY, "load_cmds: commande %s non trouve dans la hash (%s)", corename, name);
			continue;
		}
		/* cmd core diffrente de celle du network?*/
		if(strcasecmp(cmd->name, name))	HashCmd_switch(cmd, name);

		cmd->level = level;
		if(!flag) cmd->flag &= ~(CMD_DISABLE | CMD_ADMIN); /* hack! */
		else if(flag == CMD_DISABLE) cmd->flag |= CMD_DISABLE;
		else cmd->flag = flag;
		cmd->used = used;

		if(cmd->args || ChanCmd(cmd)) /* syntax load! */
		{
			strcpy(dir + 7, cmd->corename);
			strupr(dir + 7);
			if((f = fopen(dir, "r")))
			{
				fgets(cmd->syntax, SYNTAXLEN, f);
				strip_newline(cmd->syntax);
				fclose(f);
			}
			else Debug(W_TTY, "load_cmds: fichier syntax '%s' non trouv : %s", dir, strerror(errno));
		}
		i++;
	}
	fclose(fp);
	return i;
}

int write_files(aNick *nick, aChan *chan, int parc, char **parv)
{
	anUser *user;
	int i;

	for(i = 0;i < CHANHASHSIZE;i++) for(chan = chan_tab[i];chan;chan = chan->next) write_achan(chan);
	for(i = 0;i < USERHASHSIZE;i++) for(user = user_tab[i];user;user = user->next) write_anuser(user);
	write_userlist();
	write_chanlist();
#ifdef USE_WELCOMESERV
	write_welcome();
#endif
	write_badchans();
	write_badnicks();
	write_cmds();
#ifdef HAVE_VOTE
	write_votes();
#endif
	csntc(nick, "Les fichiers ont bien t crits.");
	return 1;
}

void append_file(const char *file, const char *ligne)
{
	FILE *fp = fopen(file, "a");
	if(fp)
	{
		fputs(ligne, fp), fputc('\n', fp);
		fclose(fp);
	}
	else Debug(W_WARN|W_MAX, "Impossible d'crire [%s] dans '%s' (%s)", ligne, file, strerror(errno));
	return;
}
